/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - standard device mapper
 *  Copyright (C) 2019,2020 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libfungw/fungw.h>
#include <load_cache/load_cache.h>
#include <libcschem/config.h>
#include <libcschem/abstract.h>
#include <libcschem/concrete.h>
#include <libcschem/engine.h>
#include <libcschem/actions_csch.h>
#include <libcschem/attrib.h>
#include <libcschem/libcschem.h>
#include <libcschem/library.h>
#include <libcschem/project.h>
#include <libcschem/util_parse.h>
#include <librnd/core/actions.h>
#include <librnd/core/compat_misc.h>

#include <plugins/sch_dialogs/quick_attr_util.h>

typedef struct {
	ldch_ctx_t devmaps;
	ldch_low_parser_t *low_parser;
	ldch_high_parser_t *high_parser;
} devmap_ctx_t; /* per view data */

typedef struct devmap_s {
	csch_attribs_t comp_attribs;
} devmap_t;

static const char devmap_cookie[] = "devmap";

/*** devmap parser ***/

static void parse_error(void *ectx, lht_node_t *n, const char *msg)
{
	csch_message(CSCH_MSG_ERROR, "devmap: parse error '%s' near %ld:%ld\n", msg, n->line, n->col);
}

static const csch_source_arg_t devmap_src = {'p', "devmap", "TODO:devmap_file"};
static ldch_data_t *devmap_parse(ldch_high_parser_t *parser, void *call_ctx, ldch_file_t *file)
{
	ldch_data_t *data;
	devmap_t *devmap;
	lht_doc_t *doc = ldch_lht_get_doc(file);

	if ((doc->root == NULL) || (doc->root->type != LHT_HASH) || (strcmp(doc->root->name, "std_devmap.v1") != 0)) {
		csch_message(CSCH_MSG_ERROR, "devmap: invalid root node in '%s'; expected 'ha:std_devmap.v1'\n", file->real_name);
		return NULL;
	}

	data = ldch_data_alloc(file, parser, sizeof(devmap_t));
	devmap = (devmap_t *)&data->payload;
	csch_attrib_init(&devmap->comp_attribs);
	if (csch_lht_parse_attribs(&devmap->comp_attribs, devmap_src, lht_dom_hash_get(doc->root, "comp_attribs"), parse_error, NULL) != 0) {
		csch_message(CSCH_MSG_ERROR, "devmap: failed to load any component attributes from '%s''\n", file->real_name);
		ldch_data_free(parser->ctx, data);
		return NULL;
	}

	return data;
}

static void devmap_free_payload(ldch_data_t *data)
{
	devmap_t *devmap = (devmap_t *)&data->payload;

	csch_attrib_uninit(&devmap->comp_attribs);
}

static char *devmap_lib_lookup(ldch_ctx_t *ctx, const char *load_name, ldch_low_parser_t *low, ldch_high_parser_t *high, void *low_call_ctx, void *high_call_ctx)
{
	csch_hook_call_ctx_t *cctx = high_call_ctx;
	fgw_obj_t *obj = ctx->user_data;
	csch_lib_it_t it;
	int len = strlen(load_name);
	char *real_name, *dmname = NULL;

	csch_lib_find_alloc(&it, NULL, &cctx->project->symlibs, cctx->project->prjdir);
	csch_lib_find_add_name(&it, load_name);
	if (((len > 7) && (memcmp(load_name + len - 7, ".devmap", 7) != 0)) || (len <= 7))
		csch_lib_find_add_name(&it, dmname = rnd_concat(load_name, ".devmap", NULL));

	real_name = csch_lib_find_next(&it);
	if (real_name != NULL)
		real_name = rnd_strdup(real_name);
	csch_lib_find_free(&it);

	free(dmname);
	return real_name;
}

/*** hooks ***/

fgw_error_t devmap_term2port(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	csch_acomp_t *comp;
	csch_cgrp_t *term, *sym;
	const char *pname;
	const char *slot;
	gds_t tmp;

	CSCH_HOOK_CONVARG(1, FGW_STR,  devmap_term2port, pname = argv[1].val.cstr);
	CSCH_HOOK_CONVARG(2, FGW_AOBJ, devmap_term2port, comp = fgw_aobj(&argv[2]));
	CSCH_HOOK_CONVARG(3, FGW_COBJ, devmap_term2port, term = fgw_cobj(&argv[3]));

	sym = term->hdr.parent;
	slot = csch_attrib_get_str(&sym->hdr.attr, "slot");
	if (slot == NULL)
		return 0;

	/* output port name is slot/input port name */
	gds_init(&tmp);
	gds_append_str(&tmp, slot);
	gds_append(&tmp, '/');
	gds_append_str(&tmp, pname);
	res->type = FGW_STR | FGW_DYN;
	res->val.str = tmp.array;
	return 0;
}


fgw_error_t devmap_symbol_joined_comp(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	csch_cgrp_t *sym;
	csch_acomp_t *comp;
	csch_hook_call_ctx_t *cctx = argv[0].val.argv0.user_call_ctx;
	static const csch_source_arg_t del_src = {'p', "devmap", "symbol_joined_comp"};


	CSCH_HOOK_CONVARG(1, FGW_COBJ, devmap_comp_update, sym = fgw_cobj(&argv[1]));
	CSCH_HOOK_CONVARG(2, FGW_AOBJ, devmap_comp_update, comp = fgw_aobj(&argv[2]));

	csch_attrib_del(&comp->hdr.attr, cctx->view_eng->eprio + CSCH_PRI_PLUGIN_NORMAL, "slot", del_src);
	return 0;
}

fgw_error_t devmap_compile_comp1(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	fgw_obj_t *obj = argv[0].val.argv0.func->obj;
	devmap_ctx_t *ctx = obj->script_data;
	csch_acomp_t *comp;
	const char *sdevmap;
	devmap_t *devmap;
	ldch_data_t *data;

	CSCH_HOOK_CONVARG(1, FGW_AOBJ, devmap_comp_update, comp = fgw_aobj(&argv[1]));
	assert(comp->hdr.type == CSCH_ATYPE_COMP);

	sdevmap = csch_attrib_get_str(&comp->hdr.attr, "devmap");
	if (sdevmap == NULL)
		return 0;

	data = ldch_load_(&ctx->devmaps, sdevmap, ctx->low_parser, ctx->high_parser, NULL, argv[0].val.argv0.user_call_ctx);
	if (data == NULL)
		return 0;
	devmap = (devmap_t *)&data->payload;

	csch_attrib_apply(&comp->hdr.attr, &devmap->comp_attribs, devmap_src);
	return 0;
}

fgw_error_t devmap_compile_port(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	csch_hook_call_ctx_t *cctx = argv[0].val.argv0.user_call_ctx;
	static const csch_source_arg_t pm_src = {'p', "devmap", "portmap"};
	csch_acomp_t *comp;
	csch_aport_t *port;
	const vts0_t *portmap;
	long n, len;

	CSCH_HOOK_CONVARG(1, FGW_AOBJ, devmap_comp_update, port = fgw_aobj(&argv[1]));
	assert(port->hdr.type == CSCH_ATYPE_PORT);
	comp = port->parent;
	if ((comp == NULL) || (comp->hdr.type != CSCH_ATYPE_COMP))
		return 0; /* ignore sheet ports */

	portmap = csch_attrib_get_arr(&comp->hdr.attr, "portmap");
	if (portmap == NULL)
		return 0;

	len = strlen(port->name);
	for(n = 0; n < portmap->used; n++) {
		char tmp[256], *ktmp = NULL, *key;
		const char *pm =  portmap->array[n], *set, *sep, *val;

		/* match left side of "->" with port name */
		if (strncmp(pm, port->name, len) != 0) continue;
		set = pm + len;
		if ((set[0] != '-') || (set[1] != '>')) continue;
		set+=2;

		/* parse the right side */
		sep = strchr(set, '=');
		if (sep == NULL) continue;
		val = sep+1;
		if (sep - set < sizeof(tmp)-2) {
			strncpy(tmp, set, sep-set);
			tmp[sep-set] = '\0';
			key = tmp;
		}
		else
			key = ktmp = rnd_strndup(set, sep - set);
		csch_attrib_set(&port->hdr.attr, cctx->view_eng->eprio + CSCH_PRI_PLUGIN_NORMAL, key, val, pm_src, NULL);
		free(ktmp);
	}


	return 0;
}


static int on_load(fgw_obj_t *obj, const char *filename, const char *opts)
{
	devmap_ctx_t *ctx;

	fgw_func_reg(obj, "terminal_name_to_port_name", devmap_term2port);
	fgw_func_reg(obj, "symbol_joined_component", devmap_symbol_joined_comp);
	fgw_func_reg(obj, "compile_component1", devmap_compile_comp1);
	fgw_func_reg(obj, "compile_port", devmap_compile_port);

	/* initialize view-local cache */
	obj->script_data = ctx = calloc(sizeof(devmap_ctx_t), 1);
	ldch_init(&ctx->devmaps);
	ctx->devmaps.load_name_to_real_name = devmap_lib_lookup;
	ctx->devmaps.user_data = obj;
	ctx->low_parser = ldch_lht_reg_low_parser(&ctx->devmaps);
	ctx->high_parser = ldch_reg_high_parser(&ctx->devmaps, "devmap");
	ctx->high_parser->parse = devmap_parse;
	ctx->high_parser->free_payload = devmap_free_payload;

	return 0;
}

static int on_unload(fgw_obj_t *obj)
{
	devmap_ctx_t *ctx = obj->script_data;
	ldch_uninit(&ctx->devmaps);
	free(ctx);
}


static const fgw_eng_t fgw_std_devmap_eng = {
	"std_devmap",
	csch_c_call_script,
	NULL,
	on_load,
	on_unload
};


const char csch_acts_quick_attr_devmap[] = "quick_attr_devmap(objptr)";
const char csch_acth_quick_attr_devmap[] = "Qucik Attribute Edit for devmap using the devmap library";
fgw_error_t csch_act_quick_attr_devmap(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_hidlib_t *hidlib = RND_ACT_HIDLIB;
	csch_sheet_t *sheet = (csch_sheet_t *)hidlib;
	csch_cgrp_t *grp;
	fgw_arg_t ares, args[3];
	int ret;

	QUICK_ATTR_GET_GRP(grp, "quick_attr_devmap");

	args[1].type = FGW_STR;
	args[1].val.cstr = "devmap";
	args[2].type = FGW_STR;
	args[2].val.cstr = "sheet";

	ret = rnd_actionv_bin(&sheet->hidlib, "librarydialog", &ares, 3, args);
	if (ret != 0) {
	
	}


	RND_ACT_IRES(1);
	return 0;
}

static rnd_action_t devmap_action_list[] = {
	{"quick_attr_devmap", csch_act_quick_attr_devmap, csch_acth_quick_attr_devmap, csch_acts_quick_attr_devmap}
};


int pplg_check_ver_std_devmap(int ver_needed) { return 0; }

void pplg_uninit_std_devmap(void)
{
}

int pplg_init_std_devmap(void)
{
	fgw_eng_reg(&fgw_std_devmap_eng);

	RND_REGISTER_ACTIONS(devmap_action_list, devmap_cookie);

	return 0;
}

